/*  HISTO2.C -- Generate two-dimensioal histogram for two image bands

NAME:
    HISTO2

PURPOSE:
    HISTO2 generates a two-dimensional histogram from two channels of
    image data by counting the number of co-occurrences or all possible
    pairs of pixel values (or ranges of pixel values if bin sizes
    greater than 1 are used).  The resulting frequency array may be
    output in LAS image format and/or ASCII text format.

ARGUMENTS:

    IN		Name(s) of image(s) containing two bands to be
		processed.
    OUT		Name of file to receive LAS image-format frequency
		array.
    ODTYPE	Data type of output binary image, "I*4", "REAL", or
		"ADAPT".
    ASCII	Name of file to receive ASCII frequency array.
    RECLEN	Length of each record in ASCII output array.
    FIELDSZ	Width of each field in ASCII output array.
    NBIN	Dimensions of frequency table -- number of columns and
		rows, corresponding to first and second band,
		respectively.
    BINSIZE	Sizes of each bin for first and second bands,
		respectively.
    IBIN	Specifies how the minimum pixel value in the first bin
		is to be defined for each band:  "ZERO", "MINIMUM",
		"NEGMAX", or "VALUE".
    IVALUE	Minimum value in first bin for each band, if IBIN =
		"VALUE".
    ORIGIN	Location of histogram origin:  "TOP" or "BOTTOM".
    SCALING	Scaling to apply to frequency counts:  "LINEAR" or
		"LOG".
    LOGZERO	Value to use as logarithm of zero (SCALING = "LOG"
		only). 

FILE REFERENCES:

    IN1		(read)		Input image band(s).
    IN2		(read)		Second input image band, if needed.
    OUT		(create)	Output two-dimensional frequency array,
				in LAS image format, including .IMG,
				.DDR and, if $HISTRY is set, .HIS files.
    ASCII	(create)	Output two-dimensional frequency array,
				in ASCII format.

PROGRAM HISTORY:                                  
    1992-04	Initial version, R. A. White, PSU/ORSER.
    1993-09	Ported to Unix.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:

    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in HISTO2.PDF.

EXTERNAL FUNCTIONS CALLED:

    hs2_getarg	Get arguments for HISTO2 from TAE interface.
    hs2_setbin	Determine sizes and limits of frequency count bins.
    hs2_tally	Tally the frequency counts.
    hs2_out	Output the frequency tables in requested format(s).

ALGORITHM DESCRIPTION:

    Get program arguments (hs2_getarg)
    For each input image
	Open file for input
	Determine data type and pixel size
    Allocate input buffers
    Determine sizes and limits of frequency-count bins (hs2_setbin)
    Allocate memory for frequency array
    Tally two-dimensional histogram (hs2_tally)
    Output frequency array as image and/or ASCII file (hs2_out)
    Generate processing completion message
    									
***************   End of Prolog for HISTO2   ***************	*/
									/*
Header Files:			       					*/

#ifdef vms
#include	<file.h>	/* Symbol definitions for open()	*/
#else
#include	<sys/file.h>
#endif
#include	<math.h>	/* Math function definitions		*/
#include	<stdio.h>	/* VMS standard I/O			*/
#include	<string.h>	/* C string function definitions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"alias.h"	/* LAS environment variable function	*/
#include	"util.h"	/* LAS utiltiy funciton definitions	*/
							    		/*
Header-File Constant Definitions Referenced:
    CMLEN	Maximum length of a file name (worgen.h).        
    E_SUCC	Success indicator for image I/O (worgen.h).
    IREAD	Code for image I/O read access (worgen.h).

Header-File Structure Definitions Referenced:
    PARBLK	TAE parameter block (tae.h).                              

Local Program Constants:						*/

#define  ABORT \
	    c_errmsg ("Aborting on fatal error", "histo2-fatal", &fatal)
#define	 DEFLZRATIO   -0.05	/* Default ratio of LOGZERO value to
				   log of maximum frequency count	*/
#define	 LOGPRECISION     3	/* Default number of decimal places for
				   ASCII output with scaling = "LOG"	*/
									/*
Local Function Prototypes:						*/
    void hs2_getarg (struct PARBLK *vblock, char *ascfile,
		     long *ascreclen, long bands[][3], float binsize[],
		     long *fieldsz, char ibcode[][8],
		     char in_hnm[][CMLEN], float ivalue[],
		     char *logflag, float *logzero, char *logzflag,
		     char minmax[], long nbands[], long nbin[],
		     char *odcode, char *origin, char out_hnm[CMLEN],
		     long *pmsgtim, long window[][4]);
    void hs2_setbin (char in_hnm[][CMLEN], long bands[][3],
		     char ibcode[][8], long idtype[], float binsize[],
		     float ivalue[], long nbin[]);
    void hs2_tally  (struct FDESC **fdesc, long nbands[],
		     long window[][4], long idtype[], long inpxsz[],
		     long nbin[], float binsize[], float ivalue[],
		     long offset[], long *pmsgtim, long *counts,
		     long *maxcount);
    void hs2_out    (char *out_hnm, char *odcode, char *ascfile,
		     long *ascreclen, long *fieldsz, long *counts,
		     long *maxcount, long *nbin, float *binsize,
		     float *ivalue, char *origin, char *logflag,
		     long *logprecision, float *logzero, char *minmax,
		     long *pmsgtim, struct PARBLK *vblock);
									/*
***************   Main Routine for HISTO2   ***************	*/

int main ()
    {

/*  Local Variable Definitions  */

    char	ascfile[CMLEN];	/* Name of output ASCII-format file	*/
    long	ascreclen;	/* Record length for ASCII-format file	*/
    long	bands[2][3];	/* Bands for each input image		*/
    float	binsize[2];	/* Size of bins for each band		*/
    long	*counts;	/* Array for frequency counts		*/
    long	fatal = -1;	/* Fatal error flag for c_errmsg	*/
    long	fieldsz;	/* Size of each field in ASCII output	*/
    struct FDESC
		*fdesc[3];	/* File descriptor(s) for image I/O	*/
    long	i;	      	/* Loop counter				*/
    char	ibcode[2][8];	/* Initial bin value code:  parm IBIN	*/
    long	idtype[2];	/* Input data type code			*/
    long	inpxsz[2];	/* Size of input pixels for each band	*/
    char	in_hnm[2][CMLEN]; /* Host names for input images	*/
    float	ivalue[2];	/* Initial bin value for each band	*/
    char	logflag;	/* Flag for logarithmic scaling		*/
    long	logprecision;	/* Default number of decimal places for
				   ASCII output with scaling = "LOG"	*/
    float	logzero;	/* Value to use as log. of zero count	*/
    char	logzflag;	/* Set TRUE if LOGZERO not defaulted	*/
    long	maxcount;	/* Maximum count for any bin		*/
    char	minmax[4];	/* TAE $MINMAX variable			*/
    char	msgtxt[200];	/* Buffer for error message text	*/
    long	nbands[3];	/* Number of bands for each input image */
    long	nbin[2];	/* Number of bins for each band		*/
    long	nonfatal = 2;	/* Warning flag for c_errmsg		*/
    char	odcode[6];	/* Output data type parmameter ODTYPE	*/
    long	offset[2];	/* Offsets into  thebuf  array		*/
    long	opt[2]={0,0};	/* Options for c_eopens()		*/
    char	origin[7];	/* Position of output array origin:
				   parameter ORIGIN			*/
    char	out_hnm[2][CMLEN]; /* Host name(s) of output file(s)	*/
    long	pmsgtim;	/* TAE $MSGTIM variable			*/
    long	rdflag = IREAD;	/* LAS flag for read access		*/
    long	stdlun;		/* Unit number for standard output	*/
    struct PARBLK
		vblock;		/* TAE parameter block			*/
    long	window[2][4];	/* Array of window specifiers for each
				   input image.				*/
    long	zero = 0;	/* Value 0				*/

/*  Initialize TAE and LAS  */
                                  
    c_inlas (&vblock, &stdlun);

/*  Get program arguments  */
                                                                    
    hs2_getarg(&vblock, ascfile, &ascreclen, bands, binsize, &fieldsz,
      ibcode, in_hnm, ivalue, &logflag, &logzero, &logzflag, minmax,
      nbands, nbin, odcode, origin, out_hnm[0], &pmsgtim, window);

/*  Open input data sets, getting input data types  */

    rdflag = IREAD;
    for (i = 0; nbands[i] > 0; i++)
	{
	idtype[i] = 0;
	if (c_eopens (&fdesc[i], in_hnm[i], bands[i], &window[i][0],
	  &window[i][1], &window[i][2], &window[i][3], &offset[i],
	  &rdflag, &idtype[i], &zero, opt) != E_SUCC)
	    {
	    sprintf (msgtxt, "Unable to open input %s", in_hnm[i]);
	    c_errmsg(msgtxt, "histo2-inopen", &nonfatal);
	    ABORT;
	    }
	if (idtype[i] == EBYTE)
	    inpxsz[i] = 1;
	else if (idtype[i] == EWORD)
	    inpxsz[i] = 2;
	else
	    inpxsz[i] = 4;     
	}
    fdesc[i] = NULL;
    if (nbands[1] == 0)
	{         
	idtype[1] = idtype[0];
	inpxsz[1] = inpxsz[0];
	bands[1][0] = bands [0][1];
	strcpy (in_hnm[1], in_hnm[0]);
	}

/*  Set up histogram bin assignments  */

    hs2_setbin (in_hnm, bands, ibcode, idtype, binsize, ivalue, nbin);

/*  Allocate space for counts array  */

    if ((counts = (long *) calloc (nbin[0] * nbin[1],
      sizeof (*counts)))  == NULL)
	{
	c_errmsg ("Unable to allocate space for frequency array",
	  "histo2-arralloc", &nonfatal);
	ABORT;
	}

/*  Tally the two-dimensional frequency distribution  */

    hs2_tally (fdesc, nbands, window, idtype, inpxsz, nbin, binsize,
      ivalue, offset, &pmsgtim, counts, &maxcount);

/*  If logarithmic scaling requested, determine value for log (0.0)  */

    if (maxcount <= 0)
	ABORT;
    if (logflag && (logzflag == 0))
	{
	logzero = DEFLZRATIO * 	log ((double) maxcount);
	}

/*  Output frequency array as image and/or ASCII file  */

    logprecision = LOGPRECISION;
    hs2_out (out_hnm[0], odcode, ascfile, &ascreclen, &fieldsz, counts,
      &maxcount, nbin, binsize, ivalue, origin, &logflag, &logprecision,
      &logzero, minmax, &pmsgtim, &vblock);

/*  Generate completion message  */

    i = (*out_hnm[0] != '\0') ? 1 : 0;
    if (*ascfile != '\0')
	{
	strcpy (out_hnm[i], ascfile);
	i++;
	}
    c_complt ("histo2", out_hnm, &i);

    return 0;
    }			/* End of Program HISTO2	*/
