/*  hs2_getarg -- Get arguments for HISTO2 from TAE interface

NAME:
    hs2_getarg()

PURPOSE:
    Function hs2_getarg is called by the two-dimensional histogram
    program, HISTO2, to get the program arguments from the TAE
    interface.

ARGUMENTS:
    The arguments for HISTO2 are described in the histo2.c header.

    Arguments for function hs2_getarg are:

    struct PARBLK
	    *vblock	(in)	TAE parameter block.
    char    *ascfile	(out)	Name of output ASCII-format file.
    long    *ascreclen	(out)	Record length for ASCII-format file.
    long    bands[][3]	(out)	Bands for each input image.
    float   binsize[]	(out)	Size of bins for each band.
    long    fieldsz[]	(out)	Size of each field in ASCII output.
    char    ibcode[][8]	(out)	Initial bin value code:  parm IBIN.
    char    in_hnm[][CMLEN]
			(out)   Host names for input images.
    float   ivalue[]	(out)	Initial bin value for each band.
    char    *logflag	(out)	Flag for logarithmic scaling.
    float   *logzero	(out)	Value to use as log. of zero count.
    char    *logzflag	(out)	Set TRUE if LOGZERO not defaulted.
    char    minmax[]	(out)	TAE $MINMAX variable.
    long    nbands[]	(out)	Number of bands for each input image.
    long    nbin[]		(out)	Number of bins for each band.
    char    *odcode	(out)	Output data type parmameter ODTYPE.
    char    *origin	(out)	Position of output array origin: 
				parameter ORIGIN.
    char    out_hnm[CMLEN]
			 (out)  Host name of output file.
    long    *pmsgtim	(out)	TAE $MSGTIM variable.
    long    window[][4]	(out)	Array of window specifiers for each
				input image.

FILE REFERENCES:
    Returns names of input and output files, in_hnm, out_hnm, and
    ascfile.

PROGRAM HISTORY:                                  
    1992-04	Initial version, R. A. White, PSU/ORSER.
    1993-09	Modified for SunOS 4.1.3, raw.
    1994-12	Added support for unix pathname syntax, raw.
    1996-04	Fixed problems in argument retrieval, raw.
    1997-12	Fixed error in imglimit for c_getin(), raw.
    2005-03	Updated for LAS 7.4, Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:

    This program runs under the TAE/LAS environment.
    This version has been tested under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in HISTO2.PDF.

ALGORITHM DESCRIPTION:

    Get input file name(s)
    Get name and data type code for image-format output file, if any
    Get name and field size of ASCII-format output file, if any
    Get parameters for histogram bin locations and sizes
    Get code specifying location of origin in output array
    Get frequency count scaling parameters
    Get TAE environemnt variables
    									
***************   End of Prolog for hs2_getarg   ***************	*/
									/*
Header Files:			       					*/

#include	<stdio.h>	/* Standard I/O	definitions		*/
#include	<string.h>	/* C string function definitions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"alias.h"	/* LAS environment variable function    */
#include	"labserv.h"	/* LAS label services functions		*/
#include	"util.h"	/* LAS utiltiy functions		*/
						   			/*
Local Program Constants:						*/

#define  ABORT \
	    c_errmsg ("Aborting on fatal error", "histo2-fatal", &fatal)

							      		/*
***************   Start of Function hs2_getarg   ***************	*/

FUNCTION void hs2_getarg(
    struct PARBLK
		*vblock,	/* TAE parameter block			*/
    char	*ascfile,	/* Name of output ASCII-format file	*/
    long	*ascreclen,	/* Record length for ASCII-format file	*/
    long	bands[][3],	/* Bands for each input image		*/
    float	binsize[],	/* Size of bins for each band		*/
    long	*fieldsz,	/* Size of each field in ASCII output	*/
    char	ibcode[][8],	/* Initial bin value code:  parm IBIN	*/
    char	in_hnm[][CMLEN], /* Host names for input images	*/
    float	ivalue[],	/* Initial bin value for each band	*/
    char	*logflag,	/* Flag for logarithmic scaling		*/
    float	*logzero,	/* Value to use as log. of zero count	*/
    char	*logzflag,	/* Set TRUE if LOGZERO not defaulted	*/
    char	minmax[],	/* TAE $MINMAX variable			*/
    long	nbands[],	/* Number of bands for each input image */
    long	nbin[],		/* Number of bins for each band		*/
    char	*odcode,	/* Output data type parmameter ODTYPE	*/
    char	*origin,	/* Position of output array origin:
	 			   parameter ORIGIN			*/
    char	out_hnm[CMLEN], /* Host name(s) of output file(s)	*/
    long	*pmsgtim,	/* TAE $MSGTIM variable			*/
    long	window[][4])	/* Array of window specifiers for each
				   input image.				*/
{
									/*
Local Variable Declarations:						*/

    long	bnd[2][MAXBND+1]; /* Expanded bands array for c_getin	*/
    long	dtype;		/* Data type for c_pxcopy()		*/
    long	dummy;		/* Dummy argument for function calls	*/
    char	*ext_ptr;	/* Pointer to ASCII filename extension	*/
    long	fatal = -1;	/* Fatal error code for c_errmsg	*/
    long	i;	      	/* Loop counter				*/
    long	imgcnt;		/* Total input images (not checked)	*/
    char	in_tnm[2][CMLEN];/* TAE name for input(s) [ignored]	*/
    char	msgtxt[200];	/* Buffer for error message text	*/
    long	nonfatal = 2;	/* Warning flag for c_errmsg		*/
    char	out_tnm[CMLEN]; /* TAE name for output [not used]	*/
    long	qflag[2];	/* Set by c_getin, but not checked  	*/
    char	scaling[7];	/* Scaling code for counts:  parameter
				   SCALING				*/
    long	totbnd;		/* Total input bands (not checked)	*/
    long	two = 2;	/* Constant passable by reference	*/
    long	zero = 0;	/* Value 0				*/

/*  Get program arguments -- begin with input file name(s)  */
                                                                    
    if (c_getin (vblock, "IN", &zero, in_tnm, in_hnm, &imgcnt, window,
      bnd, nbands, &totbnd, &two, &dummy, &qflag) != E_SUCC)
	{
	c_errmsg ("Error in parameter IN", "histo2-inparm", &nonfatal);
	ABORT;
	}
    bands[0][0] = bnd[0][0];
    bands[0][1] = bnd[0][1];
    bands[1][0] = bnd[1][0];
    if (nbands[0] == 2)
	{
	for (i = 0; i < 4; i++)
	    window[1][i] = window[0][i];
	nbands[1] = 0;
	}
    else if ((window[1][2] != window[0][2])
       || (window[1][3] != window[0][3]))
	{
	c_errmsg ("Input band window sizes don't match",
	  "histo2-badwind", &nonfatal);
	ABORT;
	}
    nbands[2] = 0;
    if ((nbands[0] + nbands[1]) != 2)
	{
	c_errmsg ("Must specify two input bands", "histo2-inband",
	  &nonfatal);
	ABORT;
	}

/*  Get name of LAS-format output file, if any, and data type code  */

    if (p_find (vblock, "OUT")->v_count > 0)                   
	{
	if (c_getout (vblock, "OUT", (void*) out_tnm, (void*) out_hnm)
	  != E_SUCC)
	    {
	    sprintf (msgtxt,
	      "Version of output file already exists -- %s", out_hnm);
	    c_errmsg (msgtxt, "histo2-dupoutfil", &nonfatal);
	    }
	strcpy(odcode, SVAL (*p_find (vblock, "ODTYPE"), 0));
	}
    else
	out_hnm[0] = '\0';

/*  Get name of ASCII-format output file, if any, and field size  */

    if (p_find (vblock, "ASCII")->v_count > 0)
	{
	strcpy (ascfile, SVAL(*p_find(vblock, "ASCII"), 0));
	c_getalias (ascfile, ascfile);
	if ((ext_ptr = strrchr (ascfile, HOST_EXT))
	  <= strrchr (ascfile, HOST_DIR_END))
	    sprintf (ascfile, "%s%c%s", ascfile, HOST_EXT, "dat");
	if (c_lsstat (ascfile, &zero) == E_SUCC)
	    {
	    sprintf (msgtxt,
	      "Version of output file already exists -- %s", out_hnm);
	    c_errmsg (msgtxt, "histo2-dupoutfil", &nonfatal);
	    }
	if (p_find (vblock, "FIELDSZ")->v_count > 0)
	    *fieldsz = IVAL (*p_find (vblock, "FIELDSZ"), 0);
	else
	    *fieldsz = 0;
	*ascreclen = IVAL (*p_find (vblock, "RECLEN"), 0);
	}
    else
	*ascfile = '\0';

/*  Get parameters for histogram bin locations and sizes  */

    dtype = EBYTE;
    for (i = 0; i < 2; i++)
	{
	nbin[i] = IVAL (*p_find (vblock, "NBIN"), i);
	binsize[i] = (float) RVAL (*p_find (vblock, "BINSIZE"), i);
	ivalue[i] = (float) RVAL (*p_find (vblock, "IVALUE"), i);
	strcpy(ibcode[i], SVAL (*p_find (vblock, "IBIN"), i));
	}
	
/*  Get code for location of origin in output frequency array  */

    strcpy(origin, SVAL (*p_find (vblock, "ORIGIN"), 0));

/*  Get frequency-count scaling parameters  */

    strcpy(scaling, SVAL (*p_find (vblock, "SCALING"), 0));
    if ((*logflag = (strcmp (scaling, "LOG") == 0)))
	{
	if ((*logzflag = p_find (vblock, "LOGZERO")->v_count) > 0)
	    *logzero = (float) RVAL (*p_find (vblock, "LOGZERO"), 0);
	strcpy (odcode, "REAL");
	}

/*  Get TAE environment variables  */

    strcpy(minmax, SVAL (*p_find (vblock, "$MINMAX"), 0));
    *pmsgtim = IVAL (*p_find (vblock, "$MSGTIM"), 0);

/*  Return to calling routine  */

    return;
    }			/* End of Function hs2_getarg	*/
