/*  hs2_out.c -- Output frequency array as image and/or ASCII file

NAME:
    hs2_out

PURPOSE:
    HISTO2 calls hs2_out to output the two-dimensional array of
    co-occurrences of values in each bin of the first input band with
    values in each bin of the second input band.  The frequency array is
    output in standard LAS image format and/or as an ASCII file,
    depending on whether filenames are entered for HISTO2 parameters OUT
    and ASCII, respectively.

ARGUMENTS:
    The arguments for HISTO2 are described in the histo2.c header.

    Arguments for hs2_out are:
	char	out_hnm[]	Host name of output image file.
	char	odcode[]	Output data type parmameter ODTYPE.
	char	ascfile[]	Name of output ASCII-format file.
	long	*ascreclen	Record length for ASCII-format file.
	long	*counts		Array for frequency counts.
	long	*maxcount	Maximum count for any bin.
	long	nbin[]		Number of bins for each band.
	float	binsize[]	Size of bins for each band.
	float	ivalue[]	Initial bin value for each band.
	char	origin[]	Position of output array origin:
				parameter ORIGIN.
	char	*logflag	Flag for logarithmic scaling.
	long	*logprecision	Default number of decimal places for
				ASCII output with scaling = "LOG".
	float	*logzero	Value to use as log. of zero count.
	char	minmax[]	TAE $MINMAX variable.
	long	*pmsgtim	TAE $MSGTIM variable.
	struct PARBLK
		*vblock		TAE parameter block (for $HISTRY).

FILE REFERENCES:

    If LAS-format output is specified, hs2_out creates files <OUT>.IMG,
    <OUT>.DDR, and, if the TAE global $HISTRY = "YES", <OUT>.HIS, where
    <OUT> is the filename entered for parameter OUT.  If ASCII format is
    specified, it is written to the file specified by parameter ASCII.

PROGRAM HISTORY:                                  
    1992-04	Initial version, R. A. White, PSU/ORSER.
    1993-09	Ported to SunOS 4.1.3., raw.
    2005-03	Updated for LAS 7.4, Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:

    This program runs under the TAE/LAS environment.
    It has been tested under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in HISTO2.PDF.

EXTERNAL FUNCTIONS CALLED:

ALGORITHM DESCRIPTION:

    If image format output requested
	Determine output data format
	Open output image file and allocate buffer space
    If ASCII format outut requested
	Determine field width and associated constants
	Open ASCII output file
    For each row of frequency array
	If logarithmic scaling was specified
	    Convert counts to their logarithms
	If image output requested
	    Write row to image file
	If ASCII output requested
	    While data remain in row
		Determine how many fields in next record for current row
		Output ASCII record
    Close output files and free buffers
    If image-format output
	Update DDR to show bin starting values and size
	Insert minimum and maximum value into BDR
	Create history label
    									
***************   End of Prolog for hs2_out   ***************	*/
									/*
Header Files:			       					*/

#ifdef vms
#include	<file.h>	/* Symbol definitions for open()	*/
#else
#include	<fcntl.h>	/* Symbol definitions for open()	*/
#endif
#include	<math.h>	/* Math function definitions		*/
#include	<stdio.h>	/* Standard I/O				*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"pixman.h"	/* LAS pixel manipulation functions	*/
#include	"typlim.h"	/* Data type maximum/minimum values	*/
#include	"util.h"	/* LAS utility functions		*/
						   			/*
Local Program Constants:						*/

#define  ABORT \
	    c_errmsg ("Aborting on fatal error", "histo2-fatal", &fatal)
									/*
External Variable Declarations:						*/
    extern long curent;		/* Current value of current entity for  
				   processing-progress message utility	*/

/****************   Main Routine for hs2_out   ***************	*/

FUNCTION void hs2_out (
    char	out_hnm[],	/* Host name of output image file	*/
    char	odcode[],	/* Output data type parmameter ODTYPE	*/
    char	ascfile[],	/* Name of output ASCII-format file	*/
    long	*ascreclen,	/* Record length for ASCII-format file	*/
    long	*fieldsz,	/* Size of each field in ASCII output	*/
    long	*counts,	/* Array for frequency counts		*/
    long	*maxcount,	/* Maximum count for any bin		*/
    long	nbin[],		/* Number of bins for each band		*/
    float	binsize[],	/* Size of bins for each band		*/
    float	ivalue[],	/* Initial bin value for each band	*/
    char	origin[],	/* Position of output array origin:
				   parameter ORIGIN			*/
    char	*logflag,	/* Flag for logarithmic scaling		*/
    long	*logprecision,	/* Default number of decimal places for
				   ASCII output with scaling = "LOG"	*/
    float	*logzero,	/* Value to use as log. of zero count	*/
    char	minmax[],	/* TAE $MINMAX variable			*/
    long	*pmsgtim,	/* TAE $MSGTIM variable			*/
    struct PARBLK
		*vblock)	/* TAE parameter block (for $HISTRY)	*/
    {

/*  Local Variable Definitions  */


    char	*ascbuf = NULL;	/* Buffer for ASCII output		*/
    char	*ascbufptr;	/* Current position in  ascbuf		*/
    long	ascfd = 0;	/* File descriptor for ASCII output	*/
    struct BDDR bddrbuf;	/* Band-specific data descriptor record	*/
    long	bufsiz;		/* Size of  thebuf  allocated by ealloc	*/
    long	col;		/* Current column of frequency array	*/
    long	*countptr;	/* Pointer into counts array		*/
    struct DDR
		ddrbuf;		/* Buffer for updating DDR		*/
    long	fatal = -1;	/* Fatal error flag for c_errmsg	*/
    struct FDESC
		*fdesc[2];	/* File descriptor(s) for image I/O	*/
    char	fieldspec[7];	/* ASCII output field specification	*/
    char	*fillstart	/* Start of blank fill for last ASCII   */
		 = NULL;	/*  record in row of frequency array	*/
    char	*fillstop = NULL;
				/* Last ascbuf location for blank fill	*/
    struct GDESC
		*gdesc;		/* Group descriptor for image I/O	*/
    long	i;		/* Loop counter				*/
    long	lastfield;	/* Last field in current ASCII record	*/
    float	*logrow = NULL;	/* Logarithms for one row of counts	*/
    long	ltype = ELONG;	/* Type code for LONG image pixels	*/
    double	maxval;		/* Maximum frequency value		*/
    short	minfield;	/* Minimum ASCII field size		*/
    double	minval;		/* Minimum frequency value		*/
    char	msgtxt[200];	/* Buffer for error message text	*/
    long	nfield = 0;	/* Number of fields per ASCII record	*/
    long	nfield_last = 0;/* Number of fields in last ASCII record
				   for each row of frequency array	*/
    long	nonfatal = 2;	/* Warning flag for c_errmsg		*/
    long	nrec = 0;	/* Number of ASCII records per row	*/
    long	odtype;		/* LAS data type code for image output	*/
    long	offset;		/* Offset into  thebuf  array		*/
    long	one = 1;	/* Constant passable by reference	*/
    long	outpxsz = 0;	/* Data size of image-format output	*/
    long	pmsgform;	/* Processing progress message format	*/
    long	rec;		/* Current ASCII record for row		*/
    long	row;		/* Current fow of frequency array	*/
    unsigned char
		*thebuf = NULL;	/* Buffer for image lines		*/
    float	value;		/* Value of current pixel if type REAL	*/
    long	wrflag = IWRITE; /* LAS flag for write access		*/
    long	zero = 0;	/* Value 0				*/

/*  If image-format output is desired, determine output data format  */

    if (*out_hnm != '\0')
	{
	if (odcode[0] == 'A')
	    {
	    if (*maxcount <= MAXUTINY)
		odtype = EBYTE;
	    else if (*maxcount <= MAXWORD)
		odtype = EWORD;
	    else
		odtype = ELONG;
	    }
	else if (odcode[0] == 'W')
	    {
	    if (*maxcount <= MAXWORD)
		odtype = EWORD;
	    else
		{
		odtype = ELONG;
		sprintf (msgtxt, "Maximum frequency count too big %s%s",
		  "for WORD data type -- output image will be type ",
		  "LONG");
		c_errmsg (msgtxt, "histo2-bigcount", &nonfatal);
		}
	    }
	else if (odcode[0] == 'L')
	    odtype = ELONG;
	else if (odcode[0] == 'R')
	    odtype = EREAL;
	if (odtype == EBYTE)
	    outpxsz = 1;
	else if (odtype == EWORD)
	    outpxsz = 2;
	else
	    outpxsz = 4;

/*  Open output image file and file group  */

	fdesc[0] = fdesc[1] = NULL;
	if (c_eopens(fdesc, (void *) out_hnm, &one, &one, &one,
	   &nbin[1], &nbin[0], &offset, &wrflag, &odtype, &zero, &zero)
	  != E_SUCC)
	    {
	    sprintf (msgtxt, "Unable to open output file %s", out_hnm);
	    c_errmsg (msgtxt, "histo2-openout", &nonfatal);
	    ABORT;
	    }
	if ((thebuf = (unsigned char *) c_ealloc(fdesc, &bufsiz))
	  == NULL) 
	    {
	    c_errmsg("Error allocating output buffers",
	      "histo2-bufalloc", &nonfatal);
	    c_eclose(fdesc);
	    ABORT;
	    }
	if ((c_egroup(fdesc, &gdesc, (void *) thebuf, &bufsiz)
	  != E_SUCC) ||(c_estep(&gdesc) != E_SUCC))
	    {
	    c_errmsg("Error opening output group", "histo2-groupopen",
	      &nonfatal);
	    c_eclose(fdesc);
	    ABORT;
	    }
	}

/*  If ASCII format output requested, determine field width  */

    if (*ascfile != '\0')
	{
	value = *maxcount;
	if (*logflag)
	    {
	    value = log (value);
	    if (*logzero < 0.0)
		{
		if (-10.0 * *logzero > value)
		    value = -10.0 * value;   /* Allow for minus sign */
		if (value < 10.0)
		    value = 10.0;	     /*  and leading zero    */
		}
	    else if (value < 1.0)
		value = 1.0;
	    value *= pow((double) 10.0,(double) (*logprecision));
	    value += 0.5;		     /* Allow for rounding   */
	    value *= 10.0;		     /*  and decimal point   */
	    }
	for (minfield = 1; value >= 1.0; value /= 10.0, minfield++);
	if (*fieldsz == 0)
	    *fieldsz = minfield;
	else if (*fieldsz < minfield)
	    {
	    sprintf (msgtxt, "Maximum frequency count too %s",
	      "big for specified ASCII field size");
	    c_errmsg (msgtxt, "histo2-badfldsz", &nonfatal);
	    ABORT;
	    }
	if (*logflag)
	    sprintf (fieldspec, "%%%ld.%ldf", *fieldsz,
	      *logprecision + (*fieldsz - minfield));
	else
	    sprintf (fieldspec, "%%%ldld", *fieldsz);

/*  Compute additional constants needed to control ASCII output  */

	nfield = *ascreclen / *fieldsz;
	nrec = (nbin[0] + nfield - 1) / nfield;
	nfield_last = nbin[0] - (nfield * (nrec - 1));
	ascbuf = (char *) malloc (*ascreclen + 1);
	fillstart = ascbuf + nfield_last * *fieldsz;
	fillstop = ascbuf + nfield * *fieldsz;
	for (ascbufptr = fillstop; ascbufptr < ascbuf + *ascreclen;
	  ascbufptr++)
	    *ascbufptr = ' ';

/*  Open ASCII format output file  */

	if ((ascfd = open (ascfile, O_WRONLY | O_CREAT | O_TRUNC, 0666))
	  < 0)
	    {
	    sprintf (msgtxt, "Unable to open output file %s", ascfile);
	    c_errmsg (msgtxt, "histo2-openout", &nonfatal);
	    c_egclse (&gdesc);
	    ABORT;
	    }
	}
        
/*  Start min/max calculations and/or processiing messages as requested	*/
                              
    if ((*out_hnm != '\0') && (minmax[0] == 'Y'))
	c_inmnmx(&odtype, &one, &minval, &maxval);
    if (*pmsgtim != 0)         
	{
	pmsgform = GEN2;
	c_inittm("HISTO2", "Wrote", "frequency array row", " ", " ",
	  &nbin[1], &zero, pmsgtim, &pmsgform);
	curent = 0;
	c_startm();                        
	}

/*  Output frequency array one row at a time  */

    if (*logflag)
	logrow = (float *) malloc (nbin[0] * outpxsz);
    for (row = 0; row < nbin[1]; row++, curent++)
	{
	if (origin[0] == 'T')
	    countptr = counts + row * nbin[0];
	else
	    countptr = counts + (nbin[1] - row - 1) * nbin[0];
	if (*logflag)
	    {
	    for (i = 0; i < nbin[0]; i++)
		if (countptr[i] > 0)
		    logrow[i] = log ((double) countptr[i]);
		else
		    logrow[i] = *logzero;
	    }

/*  Write row to image-format output, if requested  */

	if (*out_hnm != '\0')
	    {
	    if (*logflag)
	    	c_pxcopy ( (void*) logrow, thebuf + offset * outpxsz,
		  &odtype, nbin);
	    else
		c_pxconv (&ltype, &odtype, (void *) countptr,
		  thebuf + offset * outpxsz, nbin);
	    if (minmax[0] == 'Y')
		c_pxmnmx ( (void*) (thebuf + offset * outpxsz),
		  &nbin[0], &minval, &maxval, &odtype);
	    if (c_estep (&gdesc) != E_SUCC)
		{
		sprintf (msgtxt, "Error writing image line %ld",
		  row + 1);                                    
		c_errmsg (msgtxt, "histo2-imgwrt", &nonfatal);
		ABORT;
		}
	    }

/*  Write row to ASCII output file, if requested  */

	if (*ascfile != '\0')
	    {
	    for (rec = 1, col = 0, lastfield = nfield; rec <= nrec;
	      rec++)
		{
		if ((rec == nrec) && (nrec > 1))
		    lastfield = nfield_last;
		if (*logflag)
		    for (i = 0, ascbufptr = ascbuf; i < lastfield;
		      i++, col++, ascbufptr += *fieldsz)
			sprintf (ascbufptr, fieldspec, logrow[col]);
		else
		    for (i = 0, ascbufptr = ascbuf; i < lastfield;
		      i++, col++, ascbufptr += *fieldsz)
			sprintf (ascbufptr, fieldspec, countptr[col]);
		if ((rec == nrec) && (nrec > 1))
		    for (ascbufptr = fillstart; ascbufptr <= fillstop;
		      ascbufptr++)
			*ascbufptr = ' ';
		else
		    *fillstop = ' ';
		ascbuf[*ascreclen] = '\0';
		if (write (ascfd, ascbuf, *ascreclen) < 0)
		    {
		    sprintf (msgtxt,
		      "Error writing row %ld, record %ld%s", row, rec,
		      " to ASCII output file");
		    c_errmsg (msgtxt, "histo2-ascwrt", &nonfatal);
		    ABORT;
		    }
		}
	    }
	}

/*  Close output files and free buffers  */

    if (*pmsgtim > 0)
	c_stoptm ();
    if (*ascfile != '\0')
	close (ascfd);
    if (*out_hnm != '\0')
	{
	c_egclse (&gdesc);
	free (thebuf);
	}

/*  For image-format output, update data descriptor records  */

    if (*out_hnm != '\0')
	{
	if (c_getddr(out_hnm, &ddrbuf) != E_SUCC)
	    {
	    c_errmsg("Unable to read DDR", "histo2-getddr", &nonfatal);
	    ABORT;
	    }
	ddrbuf.pdist_x = binsize[0];
	ddrbuf.upleft[1] = ddrbuf.loleft[1] = ivalue[0];
	ddrbuf.upright[1] = ddrbuf.loright[1]
	  = ivalue[0] + (nbin[0] - 1) * binsize[0];
	if (origin[0] == 'T')
	    {
	    ddrbuf.pdist_y = binsize[1];
	    ddrbuf.upleft[0] = ddrbuf.upright[0] = ivalue[1];
	    ddrbuf.loleft[0] = ddrbuf.loright[0]
	      = ivalue[1] + (nbin[1] - 1) * binsize[1];
	    }
	else
	    {
	    ddrbuf.pdist_y = - binsize[1];
	    ddrbuf.upleft[0] = ddrbuf.upright[0]
	      = ivalue[1] + (nbin[1] - 1) * binsize[1];
	    ddrbuf.loleft[0] = ddrbuf.loright[0] = ivalue[1];
	    }
	ddrbuf.line_inc = ddrbuf.sample_inc = 1.0;
	ddrbuf.master_line = ddrbuf.master_sample = 1;
	ddrbuf.valid[DDPDV] = VALID;
	ddrbuf.valid[DDCCV] = VALID;
	ddrbuf.valid[DDINCV] = VALID;

/*  Write updated DDR, after first saving BDRs */

	if (c_getbdr(out_hnm, &bddrbuf, &one) != E_SUCC)
	    c_errmsg("Unable to read BDR", "histo2-getbdr", &nonfatal);
	if (c_putddr(out_hnm, &ddrbuf) != E_SUCC)
	    {
	    c_errmsg("Unable to write DDR", "histo2-putddr", &nonfatal);
	    ABORT;
	    }

/*  Insert minimum and maximum values into band-specific DDRs  */

	if (c_putbdr(out_hnm, &bddrbuf) != E_SUCC)
	    c_errmsg("Unable to write BDR", "histo2-putbdr", &nonfatal);
	if (c_svmnmx(out_hnm, &one, &minval, &maxval, minmax) != E_SUCC)
	    c_errmsg("Error writing min/max values","histo2-svmnmx",
	      &nonfatal);

/*  Create history label  */

	if (*out_hnm != '\0')
	    c_hstout (vblock, out_hnm, (void*) " ", &zero, " ");
	}

/*  Return to HISTO2  */

    return;
    }			/*  End of function hs2_out  */
