/*  hs2_setbin.c -- Set up bins for tallying frequency counts

NAME:
    hs2_setbin

PURPOSE:
    HISTO2 calls hs2_setbin to determine the number, size, and starting
    value for the histogram bins for each of the two input bands.

ARGUMENTS:
    The arguments for HISTO2 are described in the histo2.c header.

    Arguments for hs2_setbin are:

    char    in_hnm[][CMLEN]
			(in)	Host names for input images.
    long    bands[][3]	(in)	Bands for each input image.
    char    ibcode[][8] (in)	Initial bin value code:  parm IBIN.
    long    idtype[]	(in)	Input data type code.
    float   binsize[] (in/out)  Size of bins for each band.
    float   ivalue[]  (in/out)	Initial bin value for each band.
    long    nbin[]    (in/out)	Number of bins for each band.

FILE REFERENCES:

    .DDR files for input images (IN1, IN2) are read to determine the
    minimum and and maximum pixel values if these data are needed for
    computation of bin sizes and coverage.

PROGRAM HISTORY:                                  
    1992-04	Initial version, R. A. White, PSU/ORSER.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:

    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.X and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in hs2_setbin.PDF.

ALGORITHM DESCRIPTION:

    For each input image
	If data type is not BYTE or bin size must be computed
	    Try to get minimum and maximum from DDR (c_getbdr)
	    Abort if unsuccessful
	Determine lowest value in first bin
	Compute number of bins if not specified explicitly
	Compute bin size if not entered explicitly
    									
***************   End of Prolog for hs2_setbin   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Math function definitions		*/
#include	<stdio.h>	/* VMS standard I/O			*/
#include	<string.h>	/* C string function definitions	*/
#include 	"las.h"		/* Standard LAS include files		*/
							    		/*
Header-File Constant Definitions Referenced:
    CMLEN	Maximum length of a file name (worgen.h).        
    DD*		Field-specifcation keys for DDR (ddr.h).
    EBYTE	Pixel-type code for unsigned byte (worgen.h).
    EREAL	Code for floating point data type (worgen.h).
    E_SUCC	Success indicator for image I/O (worgen.h).
    VALID	Code indicating DDR entry is valid (worgen.h).		

Local Program Constants:						*/

#define  ABORT \
      c_errmsg ("Aborting on fatal error", "hs2_setbin-fatal", &fatal)
									/*
***************   Main Routine for hs2_setbin   ***************	*/

FUNCTION void hs2_setbin (
    char	in_hnm[][CMLEN], /* Host names for input images	*/
    long	bands[][3],	/* Bands for each input image		*/
    char	ibcode[][8],	/* Initial bin value code:  parm IBIN	*/
    long	idtype[],	/* Input data type code			*/
    float	binsize[],	/* Size of bins for each band		*/
    float	ivalue[],	/* Initial bin value for each band	*/
    long	nbin[])		/* Number of bins for each band		*/

    {

/*  Local Variable Definitions  */

    struct BDDR bddr[2];	/* Band-specific data descriptor record	*/
    long	fatal = -1;	/* Fatal error flag for c_errmsg	*/
    long	i;	      	/* Loop counter				*/
    char	msgtxt[200];	/* Buffer for error message text	*/
    char	needmnmx;	/* Set TRUE if pixel min/max values
				   needed for setting up bins		*/
    long	nonfatal = 2;	/* Warning flag for c_errmsg		*/

/*  Process each of the two input bands separately -- first determine
    whether pixel mininum and maximum are needed  */ 

    for (i = 0; i < 2; i++)
	{
	needmnmx = FALSE;
	if ((idtype[i] == EBYTE) && ((ibcode[i][0] == 'N')
	      || ((ibcode[i][0] == 'V') && (ivalue[i] < 0.0))))
	    {
	    sprintf (msgtxt, "%s%s%s%s",
	      "Parameters IBIN/IVAL for ",
	      (i == 0) ? "first" :"second",
	      " band imply negative value -- but 8-bit pixel",
	      " values are interpreted as positive");
	    c_errmsg (msgtxt, "hs2_setbin-bytnotneg", &nonfatal);
	    ABORT;
	    }
	else if ((ibcode[i][0] == 'M') || (ibcode[i][0] == 'N'))
	    needmnmx = TRUE;
	else if (idtype[i] != EBYTE)
	    {
	    if (binsize[i] == 0.0)
		needmnmx = TRUE;
	    else if (binsize[i] < 0.0)
		{
		c_errmsg ("BINSIZE value cannot be negative",
		  "hs2_setbin-negbinsiz", &nonfatal);
		ABORT;
		}
	    else if (nbin[i] == 0)
		needmnmx = TRUE;
	    }

/*  If minimum/maximum values are needed, get them from DDR  */

	if (needmnmx)
	    {
	    if (c_getbdr (&in_hnm[i][0], &bddr[i], &bands[i][0])
	      != E_SUCC)
		{
		sprintf (msgtxt, "Error reading DDR for input %s",
		  in_hnm[i]);
		c_errmsg (msgtxt, "hs2_setbin-badbdr", &nonfatal);
		ABORT;
		}
	    if (bddr[i].valid != VALID)
		{
		sprintf (msgtxt, "Minmum/maximum values for %s%s",
		  in_hnm[i], " not found -- run program minmax");
		c_errmsg (msgtxt, "hs2_setbin-nominmax", &nonfatal);
		ABORT;
		}
	    }

/*  Determine lowest value in first bin  */

	if (ibcode[i][0] == 'Z')
	    ivalue[i] = 0.0;
	else if (ibcode[i][0] == 'M')
	    ivalue[i] = bddr[i].minval;
	else if (ibcode[i][0] == 'N')
	    {
	    if (bddr[i].minval >=  - bddr[i].maxval)
		ivalue[i] = - bddr[i].maxval;
	    else
		{
		ivalue[i] = bddr[i].minval;
		bddr[i].maxval = - ivalue[i];
		}
	    }
	else if (idtype[i] != EREAL)
	    ivalue[i] = floor (ivalue[i]);

/*  Determine number of bins if not specified explicitly  */

	if (nbin[i] == 0)
	    {
	    if (idtype[i] == EBYTE)
		{
		if (ibcode[i][0] == 'M')
		    nbin[i] = 1 + bddr[i].maxval - bddr[i].minval;
		else
		    nbin[i] = 256 - ivalue[i];
		}
	    else
		{
		if (binsize[i] == 0.0)
		    {
		    if ((idtype[i] != EREAL) && needmnmx
		      && ((bddr[i].maxval - ivalue[i]) < 1023))
			nbin[i] = bddr[i].maxval - ivalue[i] + 1;
		    else
		        nbin[i] = 1024;
		    }
		else
		    {
		    nbin[i] =             
		      ceil ((bddr[i].maxval - ivalue[i]) / binsize[i]);
		    if (nbin[i] > 1024)
			{
			sprintf (msgtxt,"Computed %ld bins for%s%s%s%s",
			  nbin[i], (i == 0) ? " first" : "second",
			  " band -- if this many bins are actually",
			  " wanted, rerun program with NBIN set to",
			  " this value.");
			c_errmsg (msgtxt, "hs2_setbin-manybins",
			  &nonfatal);
			ABORT;
			}
		    }
		}
	    }

/*  Determine bin size if not entered explicitly  */

	if (binsize[i] == 0.0)
	    {
	    if (idtype[i] == EBYTE)
		binsize[i] = 1;
	    else if (idtype[i] == EREAL)
		binsize[i]
		  = 1.0000001 * (bddr[i].maxval - ivalue[i]) / nbin[i];
	    else
		binsize[i]
		  = ceil ((bddr[i].maxval - ivalue[i] + 1.0) / nbin[i]);
	    if ((binsize[i] != 1.0) || (idtype[i] != EBYTE))
		printf ("\nBin size for %s band set to %f\n",
		  (i == 0) ? "first" : "second", binsize[i]);
	    }
	}

/*  Return to HISTO2  */

    return;
    }			/* End of Program hs2_setbin */
