/*  hs2_tally.c -- Set up bins for tallying frequency counts

NAME:
    hs2_tally

PURPOSE:
    HISTO2 calls hs2_tally to tabulate the two-dimensional array of
    co-occurrences of values in each bin of the first input band with
    values in each bin of the second input band.

ARGUMENTS:
    The arguments for HISTO2 are described in the histo2.c header.

    Arguments for hs2_tally are:

	struct FDESC
		**fdesc		Input image file descriptor(s).
	long	nbands[]	Number of bands for each input image.
	long	window[][4]	Window specifiers for input images.
	long	idtype[]	Input data type codes.
	long	inpxsz[]	Size of input pixels for each band.
	long	nbin[]		Number of bins for each band.
	float	binsize[]	Size of bins for each band.
	float	ivalue[]	Initial bin value for each band.
	long	offset[]	Offsets into  thebuf  array.
	long	*pmsgtim	TAE $MSGTIM variable.
	long	*counts		Array to receive frequency counts,
				nbin[0] by nbin[1].
	long	*maxcount	Maximum count for any bin.

FILE REFERENCES:

    The input image bands specified by parameters IN1 and IN2, which
    have already been opened by HISTO2, are read to determine the values
    for each band of each pixel within the specified input window.

PROGRAM HISTORY:                                  
    1992-04	Initial version, R. A. White, PSU/ORSER.
    1994-12	Fixed malloc() usage for STDC under Solaris, raw.
    2005-03	Updated for LAS 7.4 and Linux, raw.

COMPUTER HARDWARE AND/OR SOFTWARE LIMITATIONS:

    This program runs under the TAE/LAS environment.
    It has been tested only under Solaris 2.x and Linux 2.4.

PROJECT:
    PSU/ORSER

INFORMATIONAL AND ERROR MESSAGES:

    These are described in HISTO2.PDF.

ALGORITHM DESCRIPTION:

    Allocate memory for input arrays
    For each input line
	Read data for both bands
	For each band
	    If band is not type REAL
		Convert data to type REAL
	For each pixel on line
	    If both values fall within bins
		Increment frequency count for this pair of pixel values
		Update maximum count for any bin
	    Else
		Update count of pairs outside frequency array
    Close inputs and free input buffers
    Report maximum count and out-of-bin counts, if any
    									
***************   End of Prolog for hs2_tally   ***************	*/
									/*
Header Files:			       					*/

#include	<math.h>	/* Math function definitions		*/
#include	<stdio.h>	/* VMS standard I/O			*/
#include	<string.h>	/* C string function definitions	*/
#include 	"las.h"		/* Standard LAS include files		*/
#include	"pixman.h"	/* LAS pixel manipulation functions	*/
#include	"util.h"	/* LAS utility function definitions	*/
						   			/*
Local Program Constants:						*/

#define  ABORT \
	    c_errmsg ("Aborting on fatal error", "histo2-fatal", &fatal)
									/*
External Variable Declarations:						*/
    extern long curent;		/* Current value of current entity for  
				   processing-progress message utility	*/
									/*
***************   Main Routine for HISTO2   ***************	*/

FUNCTION void hs2_tally (
    struct FDESC
		**fdesc,	/* Input image file descriptor(s)	*/
    long	nbands[],	/* Number of bands for each input image */
    long	window[][4],	/* Window specifiers for input image	*/
    long	idtype[],	/* Input data type codes		*/
    long	inpxsz[],	/* Size of input pixels for each band	*/
    long	nbin[],		/* Number of bins for each band		*/
    float	binsize[],	/* Size of bins for each band		*/
    float	ivalue[],	/* Initial bin value for each band	*/
    long	offset[],	/* Offsets into  thebuf  array		*/
    long	*pmsgtim,	/* TAE $MSGTIM variable			*/
    long	*counts,	/* Array to receive frequency counts	*/
    long	*maxcount)	/* Maximum count for any bin		*/
    {

/*  Local Variable Definitions  */

    long	bufsiz;		/* Size of  thebuf  allocated by ealloc	*/
    long	col;		/* Current column of frequency array	*/
    long	fatal = -1;	/* Fatal error flag for c_errmsg	*/
    struct GDESC
		*gdesc;		/* Group descriptor for image I/O	*/
    long	highcount[2];	/* Number of pixel values above top bin
				   for each band			*/
    long	i;	      	/* Loop counter				*/
    long	*indexarr;	/* Array of bin indices for one line	*/
    long	*indexptr;	/* Current location in indexarr		*/
    long	*indexstop;	/* Test value for  indexptr  loop	*/
    long	line;		/* Current line of input image		*/
    long	lowcount[2];	/* Number of pixel values below bottom
				   bin for each band			*/
    char	msgtxt[200];	/* Buffer for error message text	*/
    long	nlin;		/* Number of lines in each input band	*/
    long	nonfatal = 2;	/* Warning flag for c_errmsg		*/
    long	nsin;		/* Number of samples in each input line	*/
    long	outcnt;		/* Total pixels outside bins		*/
    char	outflag;	/* Flag set true if pixel value outside
				   bin limits				*/
    long	pmsgform;	/* Processing progress message format	*/
    float	*rbuf;		/* Buffer for pixels converted to real	*/
    long	row = 0;	/* Current row of frequency array	*/
    float	*rptr;		/* Pointer into buffer for real data	*/
    float	*rptrstop;	/* Last value of  rptr  for input line	*/
    long	rtype = EREAL;	/* Type code for real data		*/
    unsigned char
		*thebuf;	/* Buffer for image lines		*/
    long	zero = 0;	/* Value 0				*/

/*  Allocate space for reformatted-input arrays  */

    nlin = window[0][2];
    nsin = window[0][3];
    indexarr = (long *) calloc (2 * nsin, sizeof (*indexarr));
    if ((idtype[0] != EREAL) || (idtype[1] != EREAL))
	rbuf = (float *) calloc (nsin, sizeof (*rbuf));
    else
	rbuf = (float *) indexarr;
    if ((indexarr == NULL) || (rbuf == NULL))
	{
	c_errmsg ("Unable to allocate space for frequency array",
	  "histo2-arralloc", &nonfatal);
	ABORT;
	}

/*  Allocate buffer memory for input group  */

    if ((thebuf = (unsigned char *) c_ealloc(fdesc, &bufsiz)) == NULL)
	{
	c_errmsg ("Error allocating input buffers", "histo2-bufalloc",
	  &nonfatal);
	for (i = 0; nbands[i] > 0; i++)
	    c_eclose (&fdesc[i]);
	ABORT;
	}

/*  Define processing group for input file(s)  */

    if (c_egroup(fdesc, &gdesc, thebuf, &bufsiz) != E_SUCC)
	{                                     
	c_errmsg("Error opening input group", "histo2-groupopen",
	  &nonfatal);
	for (i = 0; nbands[i] != 0; i++)
	    c_eclose(&fdesc[i]);
	ABORT;
	}                 

/*  Start processing progress messages if requested  */

    if (*pmsgtim != 0)         
	{
	pmsgform = GEN2;
	c_inittm("HISTO2", "Processing", "input line", "", " ", &nlin,
	  &zero, pmsgtim, &pmsgform);
	curent = 1;
	c_startm();
	}

/*  Process one input line at a time  */

    *maxcount = 0;
    for (i = 0; i < 2; i++)
	highcount[i] = lowcount[i] = 0;
    indexstop = indexarr + nsin;
    for (line = 1; line <= nlin; line++, curent++)
	{

/*  Read current line for both input bands  */

	if (c_estep(&gdesc) != E_SUCC)
	    {
	    sprintf (msgtxt, "Error reading input line %ld", line);
	    c_errmsg (msgtxt, "histo2-rderr", &nonfatal);
	    c_egclse (&gdesc);
	    ABORT;
	    }

/*  Convert integer input data to real to simplify computations  */

	for (i = 0; i < 2; i++)
	    {
	    indexptr = indexarr + i * nsin;
	    if (idtype[i] != EREAL)
		{
		c_pxconv (&idtype[i], &rtype,
		  thebuf + (inpxsz[i] * offset[i]), (void *) rbuf,
		  &nsin);
		rptr = rbuf;
		}
	    else
		rptr = (float *) (thebuf + inpxsz[i] * offset[i]);
	    rptrstop = rptr + nsin;
	    for ( ; rptr < rptrstop; rptr++)
		*(indexptr++)
		  = floor ((*rptr - ivalue[i]) / binsize[i]);
	    }

/*  Increment counts in two-dimensional histogram array  */

	indexptr = indexarr;           
	for ( ; indexptr < indexstop; indexptr++)
	    {
	    outflag = FALSE;
	    if ((col = *indexptr) < 0)
		{
		outflag = TRUE;
		lowcount[0]++;
		}
	    else if (col >= nbin[0])
		{
		outflag = TRUE;
		highcount[0]++;
		}
	    else if ((row = *(indexptr + nsin)) < 0)
		{
		outflag = TRUE;
		lowcount[1]++;
		}
	    else if (row >= nbin[1])
		{
		outflag = TRUE;
		highcount[1]++;
		}
	    if (!outflag)
		if (++counts[col + nbin[0] * row] > *maxcount)
		    (*maxcount)++;
	    }
	}
    if (*pmsgtim > 0)
	c_stoptm();

/*  Report maximum count, and out-of-bin counts, if any  */

    printf ("\nMaximum frequency count is %ld, from %ld pixels %s\n",
      *maxcount, nsin * nlin, "processed");
    outcnt = lowcount[0] + lowcount[1] + highcount[0] + highcount[1];
    if (outcnt > 0)
	{
	printf(
	  "%ld pixels fell outside range of bins, as follows --\n",
	  outcnt );
	printf (
	  "    First band:   %ld below botton bin, %ld above top bin\n",
	  lowcount[0], highcount[0]);
	printf( "    Second band:  %ld below botton bin, %ld %s\n\n",
	  lowcount[1], highcount[1], "above top bin" );
	}

/*  Close inputs and free temporary storge  */

    c_egclse (&gdesc);
    free (thebuf);
    free (indexarr);
    if ((idtype[0] != EREAL) || (idtype[1] != EREAL))
	free (rbuf);

/*  Return to HISTO2  */

    return;
    }			/*  End of function hs2_tally  */
